/*
 *	main.c
 *
 *	ATARI VCS/EMU
 *
 *	* Tue Mar 15 20:31:00 JST 2005 Naoyuki Sawa
 *	- 쐬JnB
 */
#include "app.h"
//#define TEST
#ifdef TEST
void test();
#endif /*TEST*/

/****************************************************************************
 *	
 ****************************************************************************/

#define extern
#include "work.h"
#undef extern

/****************************************************************************
 *	main
 ****************************************************************************/

int
main()
{
	/* B */
	app_init();

	/* C[vB */
#ifndef TEST
	app_main();
#else /*TEST*/
	test();
#endif /*TEST*/

	return 0;
}

/****************************************************************************
 *	app_init
 ****************************************************************************/

void
app_init()
{
	/* ʓIȏB */
	pceLCDDispStop();
	pceLCDSetBuffer(vbuff);
	pceAppSetProcPeriod(1000 / FPS);
	pcePadSetTrigMode(PP_MODE_SINGLE);
	pceLCDDispStart();

	/* `B */
	surface.w = DISP_X;
	surface.h = DISP_Y;
	surface.vbuff = vbuff;
	render_init_2d(&render, &surface, 0, 0, 0, 0);

#if defined(TRACE_ON) && (TRACE_ON)
	/* WG[o͂J܂B */
	stderr_open(-1/*WG[o̓j^ڑ҂()*/);
#endif /*TRACE_ON*/

	/* NbNEEFCgؑւ܂B */
	turbo(TURBO);

	/* ؑցE16K\܂B */
	ldirect_init(LDIRECT_INTERVAL);

	/* N[Abv֐o^B */
	atexit(app_exit);
}

/****************************************************************************
 *	app_exit
 ****************************************************************************/

void
app_exit()
{
	/* ؑցE16K\܂B */
	ldirect_free();
}

/****************************************************************************
 *	schedule
 ****************************************************************************/

void
schedule()
{
#ifdef QUICK_REPEAT
	static unsigned char rep[8]; /* I[gs[gJE^ */
#endif /*QUICK_REPEAT*/

	int x;
	int y;

	pceLCDTrans();		/* ʍXV */
	yield();		/* VXe */
	joy = pcePadGet();	/* pbhԍXV */

#ifdef QUICK_REPEAT
	/* I[gs[gB */
	for(x = 0; x < 8; x++) {
		y = 1 << x; /* PAD_* */
		if(QUICK_REPEAT_PADS & y) {
			if(joy & y) {
				if(rep[x] != 0xff) rep[x]++;
				if(rep[x] >= QUICK_REPEAT_DELAY) joy |= y << 8; /* TRG_* */
			} else {
				rep[x] = 0;
			}
		}
	}
#endif QUICK_REPEAT

	now++;	/* O[o^C}XV */
}

/****************************************************************************
 *	err_disp
 ****************************************************************************/

void
err_disp(const char* fmt, ...)
{
	int len, x, y;
	char str[100]; /* ꂮ炢Ώ[H */
	va_list ap;

	va_start(ap, fmt);
	len = vsprintf(str, fmt, ap);
	va_end(ap);

	x = (DISP_X - 5 * len) / 2;
	y = (DISP_Y - 10) / 2;
	memset(vbuff, 0, sizeof vbuff);
	render_string(&render, x, y, str, 0x80, 15);
	for(;;) {
		schedule();
		if(joy & (TRG_A | TRG_B | TRG_START | TRG_SELECT)) break;
	}
}

/****************************************************************************
 *	file_init
 ****************************************************************************/

static int file_item_compare(const void* _a, const void* _b) {
	FILEITEM* a = (FILEITEM*)_a;
	FILEITEM* b = (FILEITEM*)_b;
	return strcmp(a->fname, b->fname);
}

static void file_sort() {
	qsort(file_list, file_num, sizeof(FILEITEM), file_item_compare);
}

void
file_init()
{
	int len;
	FILEINFO fi;
	FILEITEM* item;

	/* ܂NA܂B */
	memset(file_list, 0, sizeof file_list);

	/* ROM܂B */
	file_num = 0;
	item = file_list;
	pceFileFindOpen(&fi);
	while(file_num < MAX_FILE) {
		if(!pceFileFindNext(&fi)) {
			break;
		}
		len = strlen(fi.filename);
		if((len < 1 + 3) || (len > MAXFILENAME)) {
			continue;
		}
		if(strcasecmp(&fi.filename[len - (1 + 3)], ".a26") != 0) {
			continue;
		}
		strcpy(item->fname, fi.filename);
		item->size = fi.length;
		file_num++;
		item++;
	}
	pceFileFindClose(&fi);

	//file_sort();
	//Ô߁AX^bN\h
	stack_switch(tbuff + sizeof tbuff, file_sort, 0, 0);
}

/****************************************************************************
 *	app_main
 ****************************************************************************/

static int
on_notify(int type, int param)
{
	switch(type) {
	/* VXej[͉܂B */
	case APPNF_SMSTART:
		pceWaveSetChAtt(0, 127); /*  */
		break;
	case APPNF_SMRESUME:
		pceWaveSetChAtt(0,   0); /* A */
		break;
	}
	return APPNR_IGNORE;
}

void
app_main()
{
	int c;
	int i;
	int y;
	int i_file;
	FILEITEM* item;

	/* t@CXg쐬B */
	file_init();
	if(!file_num) {
		err_disp("ROM(*.a26)܂");
		exit(0);
	}

	/* ʒmR[obNo^B */
	atnotify(on_notify);

	for(;;) {
		schedule();

		/* IH */
		if(joy & TRG_SELECT) {
			exit(0);
		}

		/* t@CIύXH */
		if(joy & TRG_UP) cursel--;
		if(joy & TRG_DN) cursel++;
		if(joy & TRG_LF) cursel -= MAX_DISP;
		if(joy & TRG_RI) cursel += MAX_DISP;
		if(cursel < 0                      ) cursel = 0;
		if(cursel > file_num - 1           ) cursel = file_num - 1;
		if(cursel < curtop                 ) curtop = cursel                 ;
		if(cursel > curtop + (MAX_DISP - 1)) curtop = cursel - (MAX_DISP - 1);

		/* G~[VJnH */
		if(joy & TRG_AB) {
			item = &file_list[cursel];
			if(vcs_run(item->fname, joy & TRG_A) == 0) {
				continue;
			} else {
				err_disp("sł܂ł");
				continue;
			}
		}

		/* t@CXg\B */
		surface_clear(&surface, 0);
		item = &file_list[curtop];
		y = 8;
		for(i = 0; i < MAX_DISP + 1/*ɏ镪`*/; i++) {
			i_file = curtop + i;
			if(i_file > file_num - 1) break;
			if(i_file == cursel) {
				render_rectangle_fill(&render, 0, y - 1, DISP_X, 6 + 1, 10);
				c = 0;
			} else {
				c = 10;
			}
			render_string(&render, 4 + 4 *  1, y, item->fname, 0x82, c);
			render_printf(&render, 4 + 4 * 15, y, 0x82, c, "%8d bytes", item->size);
			item++;
			y += 6;
		}

		/* ^Cg\Biɂ͂ݏot@CXgɏ㏑j */
		render_rectangle_fill(&render, 0, 0, DISP_X, 6 + 1, 15);
		render_string(&render, -2, 1,
		//       012345678901234567890123456789012
			"========= ATARI VCS/EMU =========",
			0x82, 0);
	}
}

/****************************************************************************
 *	test
 ****************************************************************************/

#ifdef TEST
// Pitfall	Spidr.F
// 6097
// 5845
// 5121
// 4992
// 4780		3557
// 4719		3489
// 4740
// 4702
// 4664
// 4616
// 4592
// 4583
// 4578 (03/22)
// 4571
// 4555
// 4488
// 4426
// 4363
void
test()
{
	int i;
	int t0;
	int t1;

	for(i = 0; i < 10; i++) {
		schedule();
	}

	if(vcs_init("pitfall_.a26")) DIE();
	//if(vcs_init("spider_f.a26")) DIE();

	ldirect_free();

	t0 = pceTimerGetCount();
	for(i = 0; i < 60; i++) {
		/* ͍XVB */
		input_update();

		/* CPUsB */
		CYCLE += VCS_CLOCK / 60;
		m6502cpu_run(CYCLE);
		NOW += CYCLE - m6502cpu.cycle;
		CYCLE = m6502cpu.cycle;

		/* ʍXV */
//		tia_reduce(vbuff, tbuff);
//		schedule();

//		/* START+SELECTŁAQ[Ivsꂽ? */
//		if((joy & PAD_START) && (joy & PAD_SELECT)) {
//			break;
//		}
	}
	t1 = pceTimerGetCount();
	die("%d", t1 - t0);
}
#endif /*TEST*/
